import copy
import numpy
from .units import Units

class Value(object):

    def __init__(self, value=None,
                 units=None, calendar=None, _ut_unit=None,
                 format=None, names=None, definition=None,
                 cfunits=None):

        # Do units first
        if cfunits is not None:
            self._Units = cfunits
        else:
            self._Units = Units(units=units, 
                                calendar=calendar,
                                _ut_unit=_ut_unit,
                                format=format,
                                names=names, 
                                definition=definition)
        #--- End: if

        if value is not None:
            self.value = value            
    #--- End: def

    def __str__(self):  
        string = []
        if hasattr(self, 'value'):
            string.append(repr(self.value))
        if hasattr(self, 'Units'):
            string.append(str(self.Units))
        return ' '.join(string)
    #--- End: def

    def __repr__(self):
        string = ['<CF %s: ' % self.__class__.__name__]
        if hasattr(self, 'value'):
            string.append('%s,' % repr(self.value))
        if hasattr(self, 'Units'):
            string.append('units=%s' % str(self.Units))
        string.append('>')
        return ''.join(string)
    #--- End: def

#    def __iter__(self):
#        return self #self.value.flat
###    d = self.dataiter.next()
###        if self.maskiter is not None and self.maskiter.next():
###            d = masked
###        return d
##
##
#    def next(self):
##
#        v = self.value.flat.next()
#        return type(self)(value=v, cfunits=self.Units)
#
#    #--- End: def

 #   def __contains__(self, item):
 #       return item in self.value

    @property
    def ndim(self):
        return self.value.ndim
    @property
    def size(self):
        return self.value.size
    @property
    def shape(self):
        return self.value.shape
    @property
    def dtype(self):
        return self.value.dtype

    def __len__(self):
        return self.value.size 

#    def __getitem__(self, index):
#
#        print self
#        return type(self)(value=self.value[index], cfunits=self.Units)

    @property
    def units(self):
        '''
'''
        try:
            return self._Units.units
        except AttributeError:
            raise AttributeError, "'%s' object has no attribute 'units'" % \
                self.__class__.__name__
    #--- End: def 

    @units.setter
    def units(self, value):
        '''
Set the `units` attribute.
'''
       
        if hasattr(self, 'value') and self.Units:
            original_units = copy.deepcopy(self.Units)
            self._Units.units = value            
            self.value = Units.conform(self.value, original_units, self._Units,
                                       inplace=True)
        else:
            self._Units.units = value            
            
    #--- End: def

    @units.deleter
    def units(self):
        '''
Delete the `units` attribute.
'''
        try:
            del self._Units.units
        except AttributeError:            
            raise AttributeError, "Can't delete '%s' attribute 'units'" % \
                self.__class__.__name__
    #--- End: def

    @property
    def Units(self):
        '''
'''
        if hasattr(self, '_Units'):
            return self._Units

        raise AttributeError, "'%s' object has no attribute 'Units'" % \
            self.__class__.__name__
    #--- End: def 

    @Units.setter
    def Units(self, units):
        '''
Set the `Units` attribute.
'''
        if hasattr(self, 'value') and self.Units:
            self.value = Units.conform(self.value, self._Units, units, 
                                       inplace=True)

        self._Units = copy.deepcopy(units)
    #--- End: def

    @Units.deleter
    def Units(self):
        '''
Delete the `Units` attribute.
''' 
        raise AttributeError, "'%s' object attribute 'Units' is not deletable" % \
            self.__class__.__name__
    #--- End: def

    @property
    def varray(self):
        return self.value
    #--- End: def

    # ----------------------------------------------------------------
    # Binary arithmetic operators and augmented arithmetic assignments
    # ----------------------------------------------------------------
    def _add_work(self, value, value_units):
        '''

Return an object for the right hand side of addition which has been
conformed to appropiate units and set the correct units on self.
'''
        if self.Units.equivalent(value_units):
            # Equivalent units
            value = Units.conform(value, value_units, self.Units,
                                  inplace=False)
            return value

        if self.Units.isreftime and value_units.istime:
            # ref_time + time = ref_time
            new_units = Units(_ut_unit=self.Units._ut_unit)
            value = Units.conform(value, value_units, new_units,
                                  inplace=False)
            return value

        if self.Units.istime and value_units.isreftime:
            # time + ref_time = ref_time
            new_units = value_units.change_reftime_units(self.Units.units)
            value = Units.conform(value, value_units, new_units, 
                                  inplace=False)
            self.override_Units(new_units)
            return value

        if self.Units and not value_units:
            # units + no_units = units
            return value

        if not self.Units and value_units:
            # no_units + units = units
            self.override_Units(value_units)
            return value
        
        # Incompatible units
        raise TypeError, "Can't add values with %s to values with %s" % \
            (repr(value_units), repr(self.Units))
    #--- End: def

    def _sub_work(self, value, value_units):
        '''

Return an object for the right hand side of subtraction which has been
conformed to appropiate units and set the correct units on self.
'''
        if self.Units.equivalent(value_units):
            # Equivalent units
            value = Units.conform(value, value_units, self.Units, inplace=False)
            if self.Units.isreftime and value_units.isreftime:
                # Ref time - ref time = time
                self.override_Units(Units(_ut_time=self.Units._ut_time))
            return value

        if self.Units.isreftime and value_units.istime:
            # ref_time - time = ref_time
            self_Units = Units(_ut_unit=self.Units._ut_unit)
            value = Units.conform(value, value_units, self_Units, inplace=False)
            return value
        
        if self.Units.istime and value_units.isreftime:
            # time - ref_time = ref_time
            new_Units = value_units.change_reftime_units(self.Units.units)
            value = Units.conform(value, value_units, new_Units, inplace=False)
            self.override_Units(new_Units)
            return value

        if self.Units and not value_units:
            # units - no_units = units
            return value

        if not self.Units and value_units:
            # no_units - units = units
            self.override_Units(value_units)
            return value

        # Incompatible units
        raise TypeError, "Can't subtract values with %s from values with %s" % \
            (repr(value_units), repr(self.Units))
    #--- End: def

    def _mul_work(self, value, value_units):
        '''

Return an object for the right hand side of multiplication which has been
conformed to appropiate units and set the correct units on self.
'''      

        if self.Units.equivalent(value_units):
            value = Units.conform(value, value_units, self.Units,
                                  inplace=False)
            new_units = self.Units ** 2
        elif value_units:
            if self.Units:
                new_units = self.Units * value_units
            else:
                new_units = value_units
        #--- End: if

        try:
            self.override_Units(new_units)
        except NameError:
            pass

        return value
    #--- End: def

    def _div_work(self, value, value_units):
        '''

Return an object for the right hand side of division which has been
conformed to appropiate units and set the correct units on self.
'''      
        if self.Units.equivalent(value_units):
            value = Units.conform(value, value_units, self.Units, 
                                  inplace=False)
            new_units = Units('1')
        elif value_units:
            if self.Units:
                new_units = self.Units / value_units
            else:
                new_units = 1.0 / value_units
        #--- End: if
                
        try:
            self.override_Units(new_units)
        except NameError:
            pass

        return value
    #--- End: def

    def __add__(self, other):
        '''
x.__add__(y) <==> x+y
'''
        value = None
        new   = self.copy(data=False)

        if isinstance(other, self.__class__):
            value = new._add_work(other.value, other.Units)
        elif isinstance(other, (float, int, long, numpy.ndarray)):
            value = other

        if value is not None:
            new.value = self.value + value
            return new

        # Still here?
        return NotImplemented 
    #--- End: def

    def __iadd__(self, other):
        '''
x.__iadd__(y) <==> x+=y
'''
        value = None

        if isinstance(other, self.__class__):
            value = self._add_work(other.value, other.Units)
        elif isinstance(other, (float, int, long, numpy.ndarray)):
            value = other

        if value is not None:
            self.value += value
            return self

        # Still here?
        return NotImplemented
    #--- End: def

    def __radd__(self, other):
        '''
x.__radd__(y) <==> y+x
'''
        return self + other
    #--- End: def

    def __sub__(self, other):
        '''
x.__sub__(y) <==> x-y
'''
        value = None
        new   = self.copy(data=False)

        if isinstance(other, self.__class__):
            value = new._add_work(other.value, other.Units)
        elif isinstance(other, (float, int, long, numpy.ndarray)):
            value = other

        if value is not None:
            new.value = self.value - value
            return new

        # Still here?
        return NotImplemented
    #--- End: def

    def __isub__(self, other):
        '''
x.__isub__(y) <==> x-=y
'''
        value = None

        if isinstance(other, self.__class__):
            value = self._add_work(other.value, other.Units)
        elif isinstance(other, (float, int, long, numpy.ndarray)):
            value = other

        if value is not None:
            self.value -= value
            return self

        # Still here?
        return NotImplemented
    #--- End: def

    def __rsub__(self, other):
        '''
x.__rsub__(y) <==> y-x  
'''
        value = None
        new   = self.copy(data=False)

        if isinstance(other, self.__class__):
            value = new._add_work(other.value, other.Units)
        elif isinstance(other, (float, int, long, numpy.ndarray)):
            value = other

        if value is not None:
            new.value = value - self.value
            return new

        # Still here?
        return NotImplemented
    #--- End: def

    def __mul__(self, other):
        '''
x.__mul__(y) <==> x*y
'''
        value = None
        new   = self.copy(data=False)

        if isinstance(other, self.__class__):
            value = new._add_work(other.value, other.Units)
        elif isinstance(other, (float, int, long, numpy.ndarray)):
            value = other

        if value is not None:
            new.value = self.value * value
            return new

        # Still here?
        return NotImplemented
    #--- End: def

    def __imul__(self, other):
        '''
x.__imul__(y) <==> x*=y
'''
        value = None

        if isinstance(other, self.__class__):
            value = self._add_work(other.value, other.Units)
        elif isinstance(other, (float, int, long, numpy.ndarray)):
            value = other

        if value is not None:
            self.value *= value
            return self

        # Still here?
        return NotImplemented       
    #--- End: def

    def __rmul__(self, other):
        '''
x.__rmul__(y) <==> y*x  
'''
        return self * other
    #--- End: def

    def __div__(self, other):
        '''
x.__div__(y) <==> x/y
'''
        value = None
        new   = self.copy(data=False)

        if isinstance(other, self.__class__):
            value = new._add_work(other.value, other.Units)
        elif isinstance(other, (float, int, long, numpy.ndarray)):
            value = other

        if value is not None:
            new.value = self.value / value
            return new

        # Still here?
        return NotImplemented
    #--- End: def

    def __idiv__(self, other):
        '''
x.__idiv___(y) <==> x/=y
'''
        value = None

        if isinstance(other, self.__class__):
            value = self._add_work(other.value, other.Units)
        elif isinstance(other, (float, int, long, numpy.ndarray)):
            value = other

        if value is not None:
            self.value /= value
            return self

        # Still here?
        return NotImplemented       
    #--- End: def

    def __rdiv__(self, other):
        '''
x.__rdiv__(y) <==> y/x  
'''
        value = None
        new   = self.copy(data=False)

        if isinstance(other, self.__class__):
            value = new._add_work(other.value, other.Units)
        elif isinstance(other, (float, int, long, numpy.ndarray)):
            value = other

        if value is not None:
            new.value = value / self.value
            return new

        # Still here?
        return NotImplemented
    #--- End: def

    def __floordiv__(self, other):
        '''
x.__floordiv__(y) <==> x//y
'''
        value = None
        new   = self.copy(data=False)

        if isinstance(other, self.__class__):
            value = new._add_work(other.value, other.Units)
        elif isinstance(other, (float, int, long, numpy.ndarray)):
            value = other

        if value is not None:
            new.value = self.value // value
            return new

        # Still here?
        return NotImplemented
    #--- End: def

    def __ifloordiv__(self, other):
        '''
x.__ifloordiv___(y) <==> x//=y
'''
        value = None

        if isinstance(other, self.__class__):
            value = self._add_work(other.value, other.Units)
        elif isinstance(other, (float, int, long, numpy.ndarray)):
            value = other

        if value is not None:
            self.value //= value
            return self

        # Still here?
        return NotImplemented  
    #--- End: def

    def __rfloordiv__(self, other):
        '''
x.__rfloordiv__(y) <==> y//x  
'''     
        value = None
        new   = self.copy(data=False)

        if isinstance(other, self.__class__):
            value = new._add_work(other.value, other.Units)
        elif isinstance(other, (float, int, long, numpy.ndarray)):
            value = other

        if value is not None:
            new.value = value // self.value
            return new

        # Still here?
        return NotImplemented
    #--- End: def

    def __truediv__(self, other):
        '''
x.__truediv__(y) <==> x/y
'''
        value = None
        new   = self.copy(data=False)

        if isinstance(other, self.__class__):
            value = new._add_work(other.value, other.Units)
        elif isinstance(other, (float, int, long, numpy.ndarray)):
            value = other

        if value is not None:
            new.value = self.value.__truediv__(value)
            return new

        # Still here?
        return NotImplemented
    #--- End: def

    def __itruediv__(self, other):
        '''
x.__itruediv___(y) <==> x/=y
'''
        value = None

        if isinstance(other, self.__class__):
            value = self._add_work(other.value, other.Units)
        elif isinstance(other, (float, int, long, numpy.ndarray)):
            value = other

        if value is not None:
            self.value.__itruediv__(value)
            return self

        # Still here?
        return NotImplemented
    #--- End: def

    def __rtruediv__(self, other):
        '''
x.__rdiv__(y) <==> y/x  
'''
        value = None
        new   = self.copy(data=False)

        if isinstance(other, self.__class__):
            value = new._add_work(other.value, other.Units)
        elif isinstance(other, (float, int, long, numpy.ndarray)):
            value = other

        if value is not None:
            new.value = value.__truediv__(self.value)
            return new

        # Still here?
        return NotImplemented
    #--- End: def

    def __pow__(self, other):
        '''
x.__pow__(y) <==> x**y
'''
        new = type(self)()

        if self.Units:
            new.Units = self.Units ** other

        new.value = self.value ** other
        return new
    #--- End: def
    
    def __ipow__(self, other):
        '''
x.__ipow__(y) <==> x**=y
'''
        if self.Units:
            self.override_Units(self.Units**other)

        self.value **= other
        return self
    #--- End: def       

    def __rpow__(self, other):
        return NotImplemented
    #--- End: def

    # ----------------------------------------------------------------
    # Unary arithmetic operators
    # ----------------------------------------------------------------
    def __pos__(self):
        '''
x.__pos__() <==> +x
'''
        return self.copy()
    #--- End: def

    def __neg__(self, other):
        '''
x.__neg__() <==> -x
'''
        new       = self.copy(data=False)
        new.value = -self.value
        return new
    #--- End: def

    def __abs__(self):
        '''
x.__abs__() <==> abs(x)
'''
        new       = self.copy(data=False)
        new.value = abs(self.value)
        return new
    #--- End: def

    def __invert__(self):
        '''
x.__invert__() <==> ~x
'''
        new       = self.copy(data=False)
        new.value = ~self.value
        return new
    #--- End: def

    # ----------------------------------------------------------------
    # Comparison operators
    # ----------------------------------------------------------------
    def __lt__(self, other):
        '''
x.__lt__(y) <==> x<y
'''
        value = None

        if isinstance(other, self.__class__):
            new   = self.copy(data=False)
            value = new._div_work(other.value, other.Units)
     
        elif isinstance(other, (float, int, long, numpy.ndarray)):
            value = other

        if value is not None:
            return self.value < value
        
        # Still here?
        raise TypeError, "Can't compare '%s' and '%s'" % \
            (self.__class__.__name__, other.__class__.__name__)
    #--- End: def

    def __le__(self, other):
        '''
x.__le__(y) <==> x<=y
'''
        value = None

        if isinstance(other, self.__class__):
            new   = self.copy(data=False)
            value = new._div_work(other.value, other.Units)
     
        elif isinstance(other, (float, int, long, numpy.ndarray)):
            value = other

        if value is not None:
            return self.value <= value
        
        # Still here?
        raise TypeError, "Can't compare '%s' and '%s'" % \
            (self.__class__.__name__, other.__class__.__name__)
    #--- End: def

    def __gt__(self, other):
        '''
x.__gt__(y) <==> x>y
'''
        value = None

        if isinstance(other, self.__class__):
            new   = self.copy(data=False)
            value = new._div_work(other.value, other.Units)
     
        elif isinstance(other, (float, int, long, numpy.ndarray)):
            value = other

        if value is not None:
            return self.value > value
        
        # Still here?
        raise TypeError, "Can't compare '%s' and '%s'" % \
            (self.__class__.__name__, other.__class__.__name__)
    #--- End: def
  
    def __ge__(self, other):
        '''
x.__ge__(y) <==> x>=y
'''
        value = None

        if isinstance(other, self.__class__):
            new   = self.copy(data=False)
            value = new._div_work(other.value, other.Units)
     
        elif isinstance(other, (float, int, long, numpy.ndarray)):
            value = other

        if value is not None:
            return self.value >= value
        
        # Still here?
        raise TypeError, "Can't compare '%s' and '%s'" % \
            (self.__class__.__name__, other.__class__.__name__)
    #--- End: def

    def __eq__(self, other):
        '''
x.__eq__(y) <==> x==y
'''
        value = None

        if isinstance(other, self.__class__):
            new   = self.copy(data=False)
            value = new._div_work(other.value, other.Units)
     
        elif isinstance(other, (float, int, long, numpy.ndarray)):
            value = other

        if value is not None:
            return self.value == value
        
        # Still here?
        raise TypeError, "Can't compare '%s' and '%s'" % \
            (self.__class__.__name__, other.__class__.__name__)
    #--- End: def

    def __ne__(self, other):
        '''
x.__ne__(y) <==> x!=y
'''
        value = None

        if isinstance(other, self.__class__):
            new   = self.copy(data=False)
            value = new._div_work(other.value, other.Units)
     
        elif isinstance(other, (float, int, long, numpy.ndarray)):
            value = other

        if value is not None:
            return self.value != value
        
        # Still here?
        raise TypeError, "Can't compare '%s' and '%s'" % \
            (self.__class__.__name__, other.__class__.__name__)
    #--- End: def

    def copy(self, data=True):
        '''
'''
        new = type(self)(cfunits=copy.deepcopy(self.Units))

        if data and hasattr(self, 'value'):
            new.value = copy.deepcopy(self.value)

        return new
    #--- End: def
    
    def override_Units(self, units):
        '''
'''
        self._Units = copy.deepcopy(units)
    #--- End: def

#--- End: class
